pub mod query;

use crate::config::DbConfig;

use sea_orm::{
    ActiveModelTrait,
    ActiveValue::{NotSet, Set},
    ConnectionTrait, Database, DbConn, EntityTrait, TransactionTrait,
};
use serde::{Deserialize, Serialize};

type Result<T> = std::result::Result<T, sea_orm::DbErr>;

#[derive(Serialize)]
pub struct FullPackage {
    #[serde(flatten)]
    entry: entity::package::Model,
    licenses: Vec<String>,
    groups: Vec<String>,
    related: Vec<(entity::PackageRelatedEnum, String)>,
    files: Vec<String>,
}

#[derive(Serialize, Deserialize)]
#[serde(tag = "type")]
pub enum RepoType {
    Regular,
    FullMirror { mirrors: Vec<String> },
}

#[derive(Serialize, Deserialize)]
pub struct NewRepo {
    distro_id: i32,
    name: String,
    description: Option<String>,
    #[serde(flatten)]
    r#type: RepoType,
}

impl NewRepo {
    pub async fn insert(self, conn: &DbConn) -> crate::Result<entity::repo::Model> {
        let txn = conn.begin().await?;

        let repo_type = match self.r#type {
            RepoType::Regular => entity::RepoType::Regular,
            RepoType::FullMirror { .. } => entity::RepoType::FullMirror,
        };

        let repo = entity::repo::ActiveModel {
            id: NotSet,
            distro_id: Set(self.distro_id),
            name: Set(self.name),
            description: Set(self.description),
            r#type: Set(repo_type),
        };
        let model = repo.insert(conn).await?;

        match self.r#type {
            RepoType::Regular => {}
            RepoType::FullMirror { mirrors } => {
                entity::RepoMirror::insert_many(mirrors.into_iter().map(|url| {
                    entity::repo_mirror::ActiveModel {
                        id: NotSet,
                        repo_id: Set(model.id),
                        url: Set(url),
                    }
                }))
                .on_empty_do_nothing()
                .exec(&txn)
                .await?;
            }
        }

        txn.commit().await?;
        Ok(model)
    }
}

pub async fn connect(conn: &DbConfig) -> crate::Result<DbConn> {
    match conn {
        DbConfig::Sqlite {
            db_dir,
            max_connections,
        } => {
            let url = format!(
                "sqlite://{}?mode=rwc",
                db_dir.join("rieter.sqlite").to_string_lossy()
            );
            let options = sea_orm::ConnectOptions::new(url)
                .max_connections(*max_connections)
                .to_owned();

            let conn = Database::connect(options).await?;

            // synchronous=NORMAL still ensures database consistency with WAL mode, as per the docs
            // https://www.sqlite.org/pragma.html#pragma_synchronous
            conn.execute_unprepared("PRAGMA journal_mode=WAL;").await?;
            conn.execute_unprepared("PRAGMA synchronous=NORMAL;")
                .await?;

            Ok(conn)
        }
        DbConfig::Postgres {
            host,
            port,
            db,
            user,
            password,
            schema,
            max_connections,
        } => {
            let mut url = format!("postgres://{}:{}@{}:{}/{}", user, password, host, port, db);

            if !schema.is_empty() {
                url = format!("{url}?currentSchema={schema}");
            }

            let options = sea_orm::ConnectOptions::new(url)
                .max_connections(*max_connections)
                .to_owned();
            Ok(Database::connect(options).await?)
        }
    }
}
