use super::{Command, SharedState};

use std::{
    path::PathBuf,
    sync::{atomic::Ordering, Arc},
};

use sea_orm::{
    ActiveModelTrait, ColumnTrait, Condition, EntityTrait, NotSet, QueryFilter, QuerySelect, Set,
};
use sea_query::Expr;
use uuid::Uuid;

#[derive(Clone)]
pub struct Handle {
    state: Arc<SharedState>,
}

impl Handle {
    pub fn new(state: &Arc<SharedState>) -> Self {
        Self {
            state: Arc::clone(state),
        }
    }

    pub fn random_file_paths<const C: usize>(&self) -> [PathBuf; C] {
        std::array::from_fn(|_| {
            let uuid: uuid::fmt::Simple = Uuid::new_v4().into();
            self.state.repos_dir.join(uuid.to_string())
        })
    }

    pub async fn register_repo(&self, repo_id: i32) -> crate::Result<()> {
        tokio::fs::create_dir(self.state.repos_dir.join(repo_id.to_string())).await?;

        let mut repos = self.state.repos.write().await;
        repos.insert(repo_id, Default::default());

        Ok(())
    }

    pub async fn get_or_create_repo(&self, distro: &str, repo: &str) -> crate::Result<i32> {
        let distro_id: Option<i32> = entity::Distro::find()
            .filter(entity::distro::Column::Name.eq(distro))
            .select_only()
            .column(entity::distro::Column::Id)
            .into_tuple()
            .one(&self.state.conn)
            .await?;

        let distro_id = if let Some(id) = distro_id {
            id
        } else {
            let new_distro = entity::distro::ActiveModel {
                id: NotSet,
                name: Set(distro.to_string()),
                description: NotSet,
            };

            new_distro.insert(&self.state.conn).await?.id
        };

        let repo_id: Option<i32> = entity::Repo::find()
            .filter(entity::repo::Column::DistroId.eq(distro_id))
            .filter(entity::repo::Column::Name.eq(repo))
            .select_only()
            .column(entity::repo::Column::Id)
            .into_tuple()
            .one(&self.state.conn)
            .await?;

        let repo_id = if let Some(id) = repo_id {
            id
        } else {
            let new_repo = entity::repo::ActiveModel {
                id: NotSet,
                distro_id: Set(distro_id),
                name: Set(repo.to_string()),
                description: NotSet,
                r#type: Set(entity::RepoType::Regular),
            };
            let id = new_repo.insert(&self.state.conn).await?.id;

            self.register_repo(id).await?;

            id
        };

        Ok(repo_id)
    }

    pub async fn get_repo(&self, distro: &str, repo: &str) -> crate::Result<Option<i32>> {
        Ok(entity::Repo::find()
            .find_also_related(entity::Distro)
            .filter(
                Condition::all()
                    .add(entity::repo::Column::Name.eq(repo))
                    .add(entity::distro::Column::Name.eq(distro)),
            )
            .one(&self.state.conn)
            .await
            .map(|res| res.map(|(repo, _)| repo.id))?)
    }

    pub async fn remove_repo(&self, repo: i32) -> crate::Result<()> {
        self.state.repos.write().await.remove(&repo);
        entity::Repo::delete_by_id(repo)
            .exec(&self.state.conn)
            .await?;
        let _ = tokio::fs::remove_dir_all(self.state.repos_dir.join(repo.to_string())).await;

        Ok(())
    }

    /// Remove all packages in the repository that have a given arch. This method marks all
    /// packages with the given architecture as "pending deletion", before performing a manual sync
    /// & removal of stale packages.
    pub async fn remove_repo_arch(&self, repo: i32, arch: &str) -> crate::Result<()> {
        entity::Package::update_many()
            .col_expr(
                entity::package::Column::State,
                Expr::value(entity::PackageState::PendingDeletion),
            )
            .filter(
                Condition::all()
                    .add(entity::package::Column::RepoId.eq(repo))
                    .add(entity::package::Column::Arch.eq(arch)),
            )
            .exec(&self.state.conn)
            .await?;

        self.queue_sync(repo).await;
        self.queue_clean().await;

        Ok(())
    }

    pub async fn queue_pkg(&self, repo: i32, path: PathBuf) {
        self.state.tx.send(Command::ParsePkg(repo, path)).unwrap();
        self.state.repos.read().await.get(&repo).inspect(|n| {
            n.0.fetch_add(1, Ordering::SeqCst);
        });
    }

    async fn queue_sync(&self, repo: i32) {
        self.state.tx.send(Command::SyncRepo(repo)).unwrap();
    }

    async fn queue_clean(&self) {
        self.state.tx.send(Command::Clean).unwrap();
    }
}
