mod pagination;

use crate::db;
use pagination::PaginatedResponse;

use axum::{
    extract::{Path, Query, State},
    http::StatusCode,
    routing::get,
    Json, Router,
};

pub fn router() -> Router<crate::Global> {
    Router::new()
        .route("/repos", get(get_repos).post(post_repo))
        .route("/repos/:id", get(get_single_repo))
        .route("/packages", get(get_packages))
        .route("/packages/:id", get(get_single_package))
}

async fn get_repos(
    State(global): State<crate::Global>,
    Query(pagination): Query<pagination::Query>,
    Query(filter): Query<db::query::repo::Filter>,
) -> crate::Result<Json<PaginatedResponse<entity::repo::Model>>> {
    let items =
        db::query::repo::page(&global.db, pagination.per_page, pagination.page - 1, filter).await?;

    Ok(Json(pagination.res(items)))
}

async fn get_single_repo(
    State(global): State<crate::Global>,
    Path(id): Path<i32>,
) -> crate::Result<Json<entity::repo::Model>> {
    let repo = db::query::repo::by_id(&global.db, id)
        .await?
        .ok_or(axum::http::StatusCode::NOT_FOUND)?;

    Ok(Json(repo))
}

async fn post_repo(
    State(global): State<crate::Global>,
    Json(repo): Json<crate::db::NewRepo>,
) -> crate::Result<(StatusCode, Json<entity::repo::Model>)> {
    let model = repo.insert(&global.db).await?;
    global.repo.register_repo(model.id).await?;

    Ok((StatusCode::CREATED, Json(model)))
}

async fn get_packages(
    State(global): State<crate::Global>,
    Query(pagination): Query<pagination::Query>,
    Query(filter): Query<db::query::package::Filter>,
) -> crate::Result<Json<PaginatedResponse<entity::package::Model>>> {
    let items =
        db::query::package::page(&global.db, pagination.per_page, pagination.page - 1, filter)
            .await?;

    Ok(Json(pagination.res(items)))
}

async fn get_single_package(
    State(global): State<crate::Global>,
    Path(id): Path<i32>,
) -> crate::Result<Json<crate::db::FullPackage>> {
    let entry = db::query::package::full(&global.db, id)
        .await?
        .ok_or(axum::http::StatusCode::NOT_FOUND)?;

    Ok(Json(entry))
}
